/*
 *
 * Corrupt a file with one of two possible channels:
 *     - BSC of parameter p (probability of bit cross-over)
 *     - QSC of parameter p (probability of symbol error, distributed uniformly among error values)
 *
 * In both cases, the notion of a "symbol size" m (default=8) is defined. Only m bits
 * out of each byte are considered significant, and possibly affected by the channel.
 *
 * The program can be used to corrupt an image (in *.pnm format) or just any
 * plain file. If the input file is an image, specifying the "-i" flag prevents
 * the header from being corrupted.
 *
 */

#include <stdio.h>
#include <math.h>
#include <time.h>
#include <stdlib.h>
#include "rand48.h"


#define DEFAULT_BER  1e-4


int  (*outsymb)(int,FILE *), (*insymb)(FILE *),
outsymb1(int, FILE *), outsymb2(int, FILE *),
insymb1(FILE *), insymb2(FILE *);
int read_pnm_header(FILE *, int *, int *, int *, int *);
int bsc_randsymb(double ber, int m);
int qsc_randsymb(double ser, int m);
int bytewt(int x);

#define LINELEN 256
#define _M_MANDATORY

void usage()
{
	fprintf(stderr,"usage: channel [-q]  [-p<prob>] [-i] [-z<length>] [-S<seed>] -m<size> infile [outfile]\n");
	fprintf(stderr,"\nFlags, parameters:\n\
-q          : q-ary mode (default is binary)\n\
-m<size>    : symbol size (no default, mandatory)\n\
-p<prob>    : bit or symbol error rate (default = %e)\n\
-i          : input file is an image in PNM format\n\
-k<length>  : skip <length> bytes\n\
-z<length>  : no input file: generate an error stream of <length> bytes\n\
-S<seed>    : PRNG seed\n\
infile      : input file to corrupt (no default--must specify)\n\
outfile     : output file (default = stdout)\n",
	DEFAULT_BER);
}
			
			
int main(int argc, char *argv[])
{
	int i,
		inb = 0,    /* input symbol size (read from pnm header, or flag m) */
		outb = 0,    /* output symbol size */
		c1, c,
		skip=0,
		qary = 0,
		noheader=1;
	
	int rows, cols,
		maxs = 255,   /* max input value */
		type;
	int n = 0,
		ns = 0,
		n1 = 0;
	FILE *infile = NULL,
		*outfile = stdout;
	char *infilename = NULL,
		*outfilename = NULL;
	double ber=DEFAULT_BER;
	int  wts[256];
	int seed=0;
	int  zeroinput = 0;
	int len = 0;
	int err;
	
	for ( i=1; i<argc; i++ ) {
		if (argv[i][0]=='-') {
			switch( argv[i][1] ) {
				case 'i':
					noheader= 0;  /* it's an image: leave the header intact */
					break;
				case 'p':
				case 'e':
					if (sscanf(argv[i]+2,"%le",&ber)!=1){
						fprintf(stderr,"Bad flag %s\n",argv[i]);
						exit(10);
					}
					break;
				case 'q':  /* q-ary channel */
					qary = 1;
					break;
					
				case 'S':
					if (sscanf(argv[i]+2,"%d",&seed)!=1){
						fprintf(stderr,"Bad flag %s\n",argv[i]);
						exit(10);
					}
					break;
				case 'm':
					if (sscanf(argv[i]+2,"%d",&inb)!=1){
						fprintf(stderr,"Bad flag %s\n",argv[i]);
						exit(10);
					}
					break;
				case 'k':
					if (sscanf(argv[i]+2,"%d",&skip)!=1){
						fprintf(stderr,"Bad flag %s\n",argv[i]);
						exit(10);
					}
					break;
				case 'z':
					zeroinput = 1;
					if (sscanf(argv[i]+2,"%d",&len)!=1){
						fprintf(stderr,"Bad flag %s\n",argv[i]);
						exit(10);
					}
					break;
				default:
					usage();
					exit(1);
			}
		}
		else {
			if ( !zeroinput && (infilename == NULL) )
				infilename = argv[i];
			else if ( outfilename == NULL )
				outfilename = argv[i];
		}
	}
	
	
	if ( infilename != NULL ) {
		if ( (infile=fopen(infilename,"rb")) == NULL ) {
			perror(infilename);
			exit(10);
		}
	}
	else if ( !zeroinput ) {
		fprintf(stderr,"Must specify input file\n");
		usage();
		exit(1);
	}
	if ( outfilename != NULL ) {
		if ( (outfile=fopen(outfilename,"wb")) == NULL ) {
			perror(outfilename);
			exit(10);
		}
	}
	else
		outfile = stdout;
	
	
	if ( !zeroinput && !noheader ) {
		if ( read_pnm_header(infile, &cols, &rows, &maxs, &type) != 0 ){
			fprintf(stderr,"bad header\n");
			exit(10);
		}
		
		for ( inb=8; (1L<<inb)<=maxs; inb++);
		outb = inb;
		
		fprintf(outfile, "P%d\n%d %d\n%d\n",type,cols,rows,maxs);
	}
	else {

		if ( !inb )  {
#ifdef _M_MANDATORY
			fprintf(stderr,"Must specify symbol size with -m flag!\n");
			exit(10);
#else
			inb=8; /* default */
#endif		
		}


		
		outb = inb;
	}
	
	if ( inb <= 8 )
		insymb = insymb1;
	else
		insymb = insymb2;
	
	if ( outb <= 8 )
		outsymb = outsymb1;
	else
		outsymb = outsymb2;
	
	for ( i=0; i<256; i++ ) wts[i] = bytewt(i);
	
	
	if ( seed == 0) {
		int t, t8;
		t = (int)time(NULL);
		t8 = t&0xff;
		seed = t^(t8<<8)^(t8<<16);
	}
	
	srand48(seed);
	
	fprintf(stderr,"Symbol size=%d  mode: %s = %8.3le seed=%d skip=%d %s\n",
			inb, qary?"q-ary SER":"binary BER",
			ber, seed, skip,
			noheader?"":"PNM image");
	
	c = 0;
	n  = 0;
	ns = 0;
	n1 = 0;
	while ( (zeroinput && n < len) ||
		   ( !zeroinput && ( c = insymb(infile) ) != EOF ) ) {
		n ++;
		if ( n < skip )
			err = 0;
		else {
			err = qary?qsc_randsymb(ber, inb):bsc_randsymb(ber, inb);
			n1 += wts[err];
			ns += (err!=0);
		}
		
		c1 = c ^ err;
		outsymb(c1, outfile);
	}
	fprintf(stderr,
"Generated %d symbols %d bits\n"
"Actual Errors-- Bits: %d BER: %8.3e  Symbols: %d SER = %8.3e\n",
			n, n*inb, n1, (double)n1/(inb*n),ns,(double)ns/n);
	
	return 0;
}

int	insymb1(FILE *inf)
{
	return getc(inf);
}

int insymb2(FILE *inf)
{
	register int c1, c2;
	
	c1=getc(inf);
	c2=getc(inf);
	
	return (c1<<8)|c2;
}

int outsymb1(int c, FILE *outf)
{
	return putc(c, outf);
}

int outsymb2(int c, FILE *outf)
{
	putc(c>>8, outf);
	return putc(c&0xff, outf);
}

int randbit(double ber)
{
	double d = drand48();
	return (d < ber);
}


int bsc_randsymb(double ber, int m)
{
	int i, b = 0;
	for ( i=0; i<m; i++ )
		b = (b<<1) + randbit(ber);
	return b;
	
}

int qsc_randsymb(double ser, int m)
{
	int q = (1<<m), s;
	int b = randbit(ser);  /* decide whether there is a symbol error or not */
	if ( b ) {
		s = 1+lrand48()%(q-1); /* uniform distribution among nonzero error values */
	}
	else
		s = 0;
	return s;
}


int bytewt(int x)
{
	int i,w=0;
	
	for ( i=0; i<8; i++ ) {
		w += (x&1);
		x >>= 1;
	}
	return w;
}


#define LINELEN 256
int read_pnm_header(FILE *infile, int *colsp, int *rowsp, int *maxsp, int *typep)
{
	int rows, cols, maxs, type;
	char line[LINELEN+1];
	
	while ( fgets(line,LINELEN,infile)!=NULL && line[0]=='#');
	if ( ( sscanf(line,"P%d",&type) != 1 )
		|| ( type != 5 && type != 6 && type != 7 ) ) {
		fprintf(stderr,"bad header: type\n");
		return -1;
	}
	while ( fgets(line,LINELEN,infile)!=NULL && line[0]=='#');
	if ( sscanf(line,"%d %d",&cols,&rows) != 2 ) {
		fprintf(stderr,"bad header: cols rows\n");
		return -1;
	}
	while ( fgets(line,LINELEN,infile)!=NULL && line[0]=='#');
	if ( sscanf(line,"%d",&maxs) != 1 ) {
		fprintf(stderr,"bad header: maxs\n");
		return -1;
	}
	
	*colsp = cols;
	*rowsp = rows;
	*maxsp = maxs;
	*typep = type;
	return 0;
}
