#include <stdio.h>
#include <stdlib.h>

#define DEF_n 255
#define _M_MANDATORY
#define _N_MANDATORY

void
usage(){
	fprintf(stdout,
"Usage: diffblock [-m<symb_size>] [-z] [-n<block_len>] [-T<threshold>] [-v] file1 file2\n");
	fprintf(stdout,
"\nFlags, parameters:\n\
-m<symb_size>  : symbol size in bits (8)\n\
-z             : assume file2 is all-zeros (file2 not specified)\n\
-n<block_len>  : block length in symbols (255)\n\
-T<threshold>  : count blocks where there are more than T errors (0)\n\
-v             : print statistics for each block\n\
file1, file2   : files to compare\n");
}

int
main(int argc, char **argv)
{
	int n=0,
	    symbCount = 0,
	    blockCount = 0,
	    symbErrCount = 0,
	    blockErrCount = 0,
	    bitErrCount = 0,
	    curBlockSymbErrCount = 0,
		nonZero = 0,
	    verbose = 0,
	    bthreshold = 0,
	    blocksOverThreshold = 0,
	    s1, s2,
	    dblock,
	    i;
	FILE *fp1, *fp2, *msgfile = stdout;
	char *fname1=NULL, *fname2=NULL;
	int bps = 0; /* bits per symbol */
	int symbmask;
	int vszero = 0; /* compare with all-zero file */

	for ( i=1; i<argc; i++ ) {
		if (argv[i][0]=='-') {
			switch( argv[i][1] ) {
				case 'n':
					if (sscanf(argv[i]+2,"%d",&n)!=1){
						fprintf(stderr,"Bad flag %s\n",argv[i]);
						exit(10);
					}
					break;
				case 'T':
					if (sscanf(argv[i]+2,"%d",&bthreshold)!=1){
						fprintf(stderr,"Bad flag %s\n",argv[i]);
						exit(10);
					}
					break;
				case 'm':
					if (sscanf(argv[i]+2,"%d",&bps)!=1){
						fprintf(stderr,"Bad flag %s\n",argv[i]);
						exit(10);
					}
					break;
				case 'z':
					vszero = 1;
					break;
				case 'v':
					verbose ++;
					break;
				default:
					usage();
					exit(1);
			}
		}
		else if ( fname1 == NULL )
			fname1 = argv[i];
		else if ( fname2 == NULL )
			fname2 = argv[i];
		else {
			usage();
			exit(10);
		}
		
	}

	if ( fname1 == NULL ) {
		usage();
		exit(10);
	}
	else if ( (fp1 = fopen(fname1,"rb")) == NULL ) {
		perror(fname1);
		exit(10);
	}
	if ( bps==0 ) {
#ifdef _M_MANDATORY
		fprintf(stderr,"Must specify symbol size with -m flag\n");
		exit(10);
#else
		bps = 8; /* default */
#endif
	}
	
	if ( n==0 ) {
#ifdef _N_MANDATORY
		fprintf(stderr,"Must specify block length with -n flag\n");
		exit(10);
#else
		n = DEF_n; /* default */
#endif
	}
	
	if ( !vszero ) {
		if ( fname2 == NULL ) 
			vszero = 1;
		else if ( (fp2 = fopen(fname2,"rb")) == NULL ) {
			perror(fname2);
			exit(10);
		}
	}

	if ( bps > 8 ) {
		fprintf(stderr,"Sorry, this utility does not support more than 8 bits/symbol\n");
		exit(10);
	}
	curBlockSymbErrCount = 0;
	symbmask = (1<<bps)-1;
	fprintf(msgfile,"File 1: %s  File 2: %s\n", fname1, vszero?"All zeros":fname2);

	fprintf(msgfile,"Bits/symbol: %d  Block length: %d  Threshold: %d\n",bps,n,bthreshold);
	for ( dblock = 0, symbCount = 0; ; symbCount++ ) {
		s1 = getc(fp1);
		if ( vszero ) s2 = 0;
		else
			s2 = getc(fp2);
		if ( s1 == EOF ) {
			if ( !vszero && (s2 != EOF) )
				fprintf(msgfile,"EOF on file 1\n");
			break;
		}
		if ( !vszero && (s2 == EOF) ) {
			if ( s1 != EOF )
				fprintf(msgfile,"EOF on file 2\n");
			break;
		}
		if ( symbCount % n == 0 ) {
			if ( verbose ) {
				if ( symbCount ) {
					printf("Block %6d  Symb errors: %4d %s\n",blockCount-1,curBlockSymbErrCount,(curBlockSymbErrCount>bthreshold)?"++":"");
				}
			}
			dblock = 0;
			blockCount ++;
			if ( curBlockSymbErrCount > bthreshold )
				blocksOverThreshold++;
			if ( curBlockSymbErrCount ) nonZero++;
			curBlockSymbErrCount = 0;
		}

		s1 &= symbmask;
		s2 &= symbmask;
		if ( s1 != s2 ) {
		    unsigned x = s1^s2;
		    while ( x ) {
		    	if ( x&1 ) bitErrCount++;
				x >>= 1;
		    }
		    symbErrCount ++;
		    if ( !dblock ) blockErrCount++;
		    dblock = 1;
		    curBlockSymbErrCount++;
		}
	}

	if ( symbCount % n ) {
		printf("Last block truncated to %d symbols\n",symbCount%n);
	}
	
	if ( curBlockSymbErrCount > bthreshold )
		blocksOverThreshold++;

	if ( verbose ) {
		if ( symbCount ) {
			printf("Block %6d  Symbol errors: %4d\n",blockCount,curBlockSymbErrCount);
		}
	}
	/*blockErrCount += dblock; */
	fprintf(msgfile,"Total symb: %8d  Blks: %6d\n",symbCount, blockCount);
	fprintf(msgfile,"Error counts Bits: %10d  Symbols: %10d  Blocks: %10d    >T: %6d\n",bitErrCount,symbErrCount, blockErrCount, blocksOverThreshold);
	fprintf(msgfile,"Error rates  Bits: %10.3e  Symbols: %10.3e  Blocks: %10.3e\n",
		(double)bitErrCount/(bps*symbCount),
		(1.0*symbErrCount)/symbCount,
		(1.0*blockErrCount)/blockCount);
	return 0;
}

